/**
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <inference_engine_error.h>
#include "inference_engine_opencv_private.h"

#include <opencv2/core/ocl.hpp>

#include <fstream>
#include <iostream>
#include <unistd.h>
#include <time.h>

namespace InferenceEngineImpl {
namespace OpenCVImpl {

InferenceOpenCV::InferenceOpenCV(std::string protoFile, std::string weightFile, 
    std::string userFile) :
    mNet(),
    mConfigFile(protoFile),
    mWeightFile(weightFile),
    mUserFile(userFile)
{
    LOGE("ENTER");
    LOGE("LEAVE");
}

InferenceOpenCV::~InferenceOpenCV()
{
    ;
}

int InferenceOpenCV::SetUserFile()
{
    std::ifstream fp(mUserFile.c_str());
    if (!fp.is_open()) {
        return INFERENCE_ENGINE_ERROR_INVALID_PATH;
    }

    std::string userListName;
    while (!fp.eof()) {
        std::getline(fp, userListName);
        if (userListName.length())
            SetUserListName(userListName);
    }

    fp.close();

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetInputTensorParam()
{
    LOGE("Not supported");
    return INFERENCE_ENGINE_ERROR_NOT_SUPPORTED;
}

int InferenceOpenCV::SetInputTensorParamInput(int width, int height, int dim, int ch)
{
    mCh = ch;
    mDim = dim;
    mInputSize = cv::Size(width, height);

    LOGE("InputSize is %d x %d\n", mInputSize.width, mInputSize.height);

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetInputTensorParamNorm(double deviation, double mean)
{
    mDeviation = deviation;
    mMean = mean;

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetInputTensorParamNode(std::string node)
{
    mInputLayer = cv::String(node);
    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetOutputTensorParam()
{
    return INFERENCE_ENGINE_ERROR_NOT_SUPPORTED;
}

int InferenceOpenCV::SetOutputTensorParamThresHold(double threshold)
{
    mThreshold = threshold;

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetOutputTensorParamNumbers(int number)
{
    mOutputNumbers = number;

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetOutputTensorParamType(int type)
{
    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetOutputTensorParamNodes(std::vector<std::string> nodes)
{
    mOutputLayer.clear();
    for (std::vector<std::string>::iterator iter = nodes.begin();
        iter != nodes.end(); ++iter) {
        mOutputLayer.push_back(cv::String(*iter));
    }

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::SetTargetDevice(inference_target_type_e type)
{
    switch (type) {
    case INFERENCE_TARGET_CPU :
        mNet.setPreferableTarget(cv::dnn::DNN_TARGET_CPU);
        break;
    case INFERENCE_TARGET_GPU :
        mNet.setPreferableTarget(cv::dnn::DNN_TARGET_OPENCL);
        break;
    case INFERENCE_TARGET_CUSTOM:
    case INFERENCE_TARGET_NONE:
    default:
        LOGE("Not supported device type [%d], Set CPU mode", (int)type);
        mNet.setPreferableTarget(cv::dnn::DNN_TARGET_CPU);
    }

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::Load()
{
    int ret = INFERENCE_ENGINE_ERROR_NONE;

    if (access(mConfigFile.c_str(), F_OK) || 
        access(mWeightFile.c_str(), F_OK)) {
        LOGE("protofile in [%s] ", mConfigFile.c_str());
        LOGE("weightFilePath in [%s] ", mWeightFile.c_str());
		return INFERENCE_ENGINE_ERROR_INVALID_PATH;
	}

    size_t userFileLength = mUserFile.length();
    if (userFileLength > 0 && access(mUserFile.c_str(), F_OK)) {
        LOGE("userFilePath in [%s] ", mUserFile.c_str());
		return INFERENCE_ENGINE_ERROR_INVALID_PATH;
	}

    // This call may be changed if OpenCV version would be upgraded
    int nPos = mWeightFile.find_last_of(".");
    std::string weightFileExt = mWeightFile.substr(nPos+1);
    LOGI("%s", weightFileExt.c_str());
    if (weightFileExt.compare("caffemodel") == 0) {
        mNet = cv::dnn::readNetFromCaffe(mConfigFile, mWeightFile);
    } else if (weightFileExt.compare("pb") == 0) {
        mNet = cv::dnn::readNetFromTensorflow(mWeightFile, mConfigFile);
    } else {
        LOGE("Not supported model file!");
    }

    if (mNet.empty()) {
        LOGE("Net is empty");
        return INFERENCE_ENGINE_ERROR_INVALID_DATA;
    }

    ret = (userFileLength > 0) ? SetUserFile() : INFERENCE_ENGINE_ERROR_NONE;
    if (ret != INFERENCE_ENGINE_ERROR_NONE)
        LOGE("Fail to read categoryList");

    return ret;
}

int InferenceOpenCV::CreateInputLayerPassage()
{
    /* Do Nothing
     * In OpenCV, don't need to call this CreateInputLayerPassage()
     * because that it can uses cv::Mat directly
     */

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::PrepareInputLayerPassage(inference_input_type_e type)
{
    /* Do Nothing
     * In OpenCV, don't need to call this PrepareInputLayerPassage()
     * because that it can uses cv::Mat directly
     */
    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::Run(cv::Mat tensor)
{
    double scaleVal = 1.0/mDeviation;
    mSourceSize = tensor.size();
    if (mInputSize == cv::Size(-1,-1)) {
        mInputSize = mSourceSize;
    }

    cv::Scalar meanScalar = cv::Scalar((float)mMean,(float)mMean, (float)mMean);

    mInputBlob = cv::dnn::blobFromImage(tensor, scaleVal, mInputSize, meanScalar, false, false);

    if (mInputBlob.empty()) {
        LOGE("Fail to get input blob");
        return INFERENCE_ENGINE_ERROR_INTERNAL;
    }

    mNet.setInput(mInputBlob, mInputLayer);

    /* foward */
    mNet.forward(mOutputProb, mOutputLayer);

    if (mOutputProb.empty()) {
        LOGE("OutputProb is empty");
        return INFERENCE_ENGINE_ERROR_INTERNAL;
    }

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::Run(std::vector<float> tensor)
{
    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::GetInferenceResult(ImageClassificationResults& results)
{
    int dims = mOutputProb[0].dims;
    for (int k = 0; k < dims; ++k) {
        LOGE("%d: %d", k, mOutputProb[0].size[k]);
    }
    LOGI("dims: %d", dims);
    cv::Mat reShapedProb = mOutputProb[0].reshape(1,1);


    int classIdx = -1;
    double classProb = 0.0;

    cv::Mat outputIdx, outputProb;

    cv::sortIdx(reShapedProb, outputIdx, CV_SORT_EVERY_ROW + CV_SORT_DESCENDING);

    for (int idx = 0; idx < mOutputNumbers; ++idx) {
        classIdx = outputIdx.at<int>(0,idx);
        int y = classIdx / reShapedProb.cols;
        int x = classIdx % reShapedProb.cols;

        classProb = (double)reShapedProb.at<float>(y,x);
        results.indices.push_back(classIdx);
        results.confidences.push_back(classProb);
        LOGI("classIdx: %d", classIdx);
        LOGI("classProb: %lf", classProb);
        LOGI("test....");
        LOGI("names: %s", mUserListName[classIdx].c_str());
        results.names.push_back(mUserListName[classIdx]);
        
    }

    results.number_of_classes = mOutputNumbers;

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::GetInferenceResult(ObjectDetectionResults& results)
{
    int dims = mOutputProb[0].dims;
    for (int k = 0; k < dims; ++k) {
        LOGE("%d: %d", k, mOutputProb[0].size[k]);
    }
    LOGI("dims: %d", dims);
    cv::Mat detectionMat(mOutputProb[0].size[2], mOutputProb[0].size[3], CV_32F, mOutputProb[0].ptr<float>());

    float confidence;
    size_t objectClass;
    int left, top, right, bottom;
    cv::Rect loc;
    for(int i = 0; i < detectionMat.rows; i++) {
        confidence = detectionMat.at<float>(i, 2);

        if(confidence > (float)mThreshold) {
            objectClass = (size_t)(detectionMat.at<float>(i, 1));

            left = static_cast<int>(detectionMat.at<float>(i, 3) * mSourceSize.width);
            top = static_cast<int>(detectionMat.at<float>(i, 4) * mSourceSize.height);
            right = static_cast<int>(detectionMat.at<float>(i, 5) * mSourceSize.width);
            bottom = static_cast<int>(detectionMat.at<float>(i, 6) * mSourceSize.height);

            loc.x = left;
            loc.y = top;
            loc.width = right - left + 1;
            loc.height = bottom - top + 1;

            //classNames[objectClass];
            LOGE("objectClass: %d", objectClass);
            LOGE("confidence:%f", confidence);
            LOGE("left:%d, top:%d, right:%d, bottom:%d", left, top, right, bottom);

            results.indices.push_back(objectClass);
            results.confidences.push_back(confidence);
            results.names.push_back(mUserListName[objectClass]);
            results.locations.push_back(loc);
        }
    }

    results.number_of_objects = results.indices.size();

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::GetInferenceResult(FaceDetectionResults& results)
{
    int dims = mOutputProb[0].dims;
    for (int k = 0; k < dims; ++k) {
        LOGE("%d: %d", k, mOutputProb[0].size[k]);
    }
    LOGI("dims: %d", dims);

    cv::Mat detectionMat(mOutputProb[0].size[2], mOutputProb[0].size[3], CV_32F, mOutputProb[0].ptr<float>());

    float confidence;
    int left, top, right, bottom;
    cv::Rect loc;
    for(int i = 0; i < detectionMat.rows; i++) {
        confidence = detectionMat.at<float>(i, 2);

        if(confidence > (float)mThreshold) {
            left = static_cast<int>(detectionMat.at<float>(i, 3) * mSourceSize.width);
            top = static_cast<int>(detectionMat.at<float>(i, 4) * mSourceSize.height);
            right = static_cast<int>(detectionMat.at<float>(i, 5) * mSourceSize.width);
            bottom = static_cast<int>(detectionMat.at<float>(i, 6) * mSourceSize.height);

            loc.x = left;
            loc.y = top;
            loc.width = right - left + 1;
            loc.height = bottom - top + 1;

            //classNames[objectClass];
            LOGE("confidence:%f", confidence);
            LOGE("left:%d, top:%d, right:%d, bottom:%d", left, top, right, bottom);

            results.confidences.push_back(confidence);
            results.locations.push_back(loc);
        }
    }

    results.number_of_faces = results.locations.size();

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::GetInferenceResult(FacialLandMarkDetectionResults& results)
{
    int dims = mOutputProb[0].dims;
    for (int k = 0; k < dims; ++k) {
        LOGE("%d: %d", k, mOutputProb[0].size[k]);
    }
    LOGI("dims: %d", dims);

    cv::Mat outputs = mOutputProb[0].reshape(1,1);

    int number_of_landmarks = (outputs.cols >> 1);

    int pos = 0;
    cv::Point pint;
    for (int idx = 0; idx < number_of_landmarks; ++idx) {
        pos = (idx << 1);
        pint.x = static_cast<int>((outputs.at<float>(0, pos) + 0.5) * mSourceSize.width);
        pint.y = static_cast<int>((outputs.at<float>(0, pos+1) + 0.5) * mSourceSize.height);

        results.locations.push_back(pint);
        LOGI("x:%d, y:%d", pint.x, pint.y);
    }

    results.number_of_landmarks = number_of_landmarks;

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::GetInferenceResult(std::vector<std::vector<int>>& dimInfo, std::vector<float*>& results)
{
    dimInfo.clear();
    results.clear();
    std::vector<int> tmpDimInfo;
    LOGE("outputProb size: %d", mOutputProb.size());
    for (std::vector<cv::Mat>::iterator iter = mOutputProb.begin();
        iter != mOutputProb.end(); ++iter) {
        tmpDimInfo.clear();
        for (int d = 0; d < (*iter).dims; ++d) {
            tmpDimInfo.push_back((*iter).size[d]);
        }

        dimInfo.push_back(tmpDimInfo);
        results.push_back((*iter).ptr<float>());
    }

    return INFERENCE_ENGINE_ERROR_NONE;
}

int InferenceOpenCV::GetNumberOfOutputs()
{
    return mOutputNumbers;
}

void InferenceOpenCV::SetUserListName(std::string userlist)
{
    mUserListName.push_back(userlist);
}

extern "C"
{
class IInferenceEngineVision* EngineVisionInit(std::string protoFile, std::string weightFile,
    std::string userFile)
{
    InferenceOpenCV *engine = new InferenceOpenCV(protoFile, weightFile, userFile);
    return engine;
}

void EngineVisionDestroy(class IInferenceEngineVision *engine)
{
    delete engine;
}

class IInferenceEngineCommon* EngineCommonInit(std::string protoFile, std::string weightFile,
    std::string userFile)
{
    InferenceOpenCV *engine = new InferenceOpenCV(protoFile, weightFile, userFile);
    return engine;
}

void EngineCommonDestroy(class IInferenceEngineCommon *engine)
{
   delete engine;
}
}
} /* OpenCVImpl */
} /* InferenceEngineImpl */
